unit DXML;

interface

uses
  DUtils;

type
TXMLItemType = (xitHeader, xitTag, xitText);

TXMLParam = record
  Name: TString;
  Value: TString;
end;

PXMLItem = ^TXMLItem;
TXMLItem = record
  Typ: TXMLItemType;
  Text: TString; // for tag it is name; for header it is ''
  Params: array of TXMLParam; // for text it is nil
  Content: array of PXMLItem; // just for tag
end;

TXML = class
private
  FHeader: TXMLItem;
  FRoot: TXMLItem;
public
  constructor Create;
  destructor Destroy; override;
  function Load(FileName: TString): Boolean;
  procedure Save(FileName: TString);
  procedure Clear;
  procedure SkipSpaces(Data: TData);
  function ReadParams(Data: TData; Item: PXMLItem; Finish: TSetOfChar): Boolean;
  function ReadTag(Data: TData; Item: PXMLItem): Boolean;
  function ReadContent(Data: TData; Item: PXMLItem; var EndTag: TString): Boolean;
  function ReadStr(Data: TData; var S: TString): Boolean;
  function GetHeader: PXMLItem;
  function GetRoot: PXMLItem;
  property Root: PXMLItem read GetRoot;
end;

procedure XMLSetEmpty(Item: PXMLItem; Typ: TXMLItemType; Text: TString);
procedure XMLAddParam(Item: PXMLItem; Name, Value: TString);
function XMLAddContent(Item: PXMLItem; Typ: TXMLItemType; Text: TString): PXMLItem;

implementation

procedure XMLSetEmpty(Item: PXMLItem; Typ: TXMLItemType; Text: TString);
begin
  Item.Typ := Typ;
  Item.Text := Text;
  SetLength(Item.Params, 0);
  SetLength(Item.Content, 0);
end;

procedure XMLAddParam(Item: PXMLItem; Name, Value: TString);
begin
  if Item.Typ = xitText then
    Exit;
  SetLength(Item.Params, Length(Item.Params) + 1);
  Item.Params[High(Item.Params)].Name := Name;
  Item.Params[High(Item.Params)].Value := Value;
end;

function XMLAddContent(Item: PXMLItem; Typ: TXMLItemType; Text: TString): PXMLItem;
begin
  if Item.Typ <> xitTag then
    Exit;
  SetLength(Item.Content, Length(Item.Content) + 1);
  New(Item.Content[High(Item.Content)]);
  Result := Item.Content[High(Item.Content)];
  Result.Typ := Typ;
  Result.Text := Text;
  SetLength(Result.Params, 0);
  SetLength(Result.Content, 0);
end;

constructor TXML.Create;
begin
  Clear;
end;

destructor TXML.Destroy;
begin
end;

function TXML.Load(FileName: TString): Boolean;
  var
    Data: TData;
begin
  Result := False;
  Data := TData.Create(FileName);
  SkipSpaces(Data);
  if Data.ReadByte <> Ord('<') then
    Exit;
  if Data.ReadByte = Ord('?') then begin
    if not ReadStr(Data, FHeader.Text) then
      Exit;
    if not ReadParams(Data, @FHeader, ['?']) then
      Exit;
    Data.ReadByte;
    if Data.ReadByte <> Ord('>') then
      Exit;
    SkipSpaces(Data);
  end;
  if Data.ReadByte <> Ord('<') then
    Exit;
  if Byte(Data.GetPtr^) = Ord('!') then
    Exit;
  if not ReadTag(Data, @FRoot) then
    Exit;
  Result := True;
end;

procedure TXML.Save(FileName: TString);
begin
end;

procedure TXML.Clear;
begin
  XMLSetEmpty(@FHeader, xitHeader, '');
  XMLSetEmpty(@FRoot, xitTag, '');
end;

procedure TXML.SkipSpaces(Data: TData);
begin
  while Byte(Data.GetPtr^) <= 32 do
    Data.ReadByte;
end;

function TXML.ReadParams(Data: TData; Item: PXMLItem; Finish: TSetOfChar): Boolean;
  var
    Name, Value: String;
begin
  Result := False;
  SkipSpaces(Data);
  while not (TChar(Data.GetPtr^) in Finish) do begin
    Name := '';
    while not (TChar(Data.GetPtr^) in [#0..#32, '=']) do
      Name := Name + TChar(Data.ReadByte);
    SkipSpaces(Data);
    if Data.ReadByte <> Ord('=') then
      Exit;
    SkipSpaces(Data);
    Value := '';
    if TChar(Data.GetPtr^) = '"' then begin
      Data.ReadByte;
      while TChar(Data.GetPtr^) <> '"' do
        Value := Value + TChar(Data.ReadByte);
      Data.ReadByte;
    end else if TChar(Data.GetPtr^) = '''' then begin
      Data.ReadByte;
      while TChar(Data.GetPtr^) <> '''' do
        Value := Value + TChar(Data.ReadByte);
      Data.ReadByte;
    end else
      Exit;
    XMLAddParam(Item, Name, Value);
    SkipSpaces(Data);
  end;
  Result := True;
end;

function TXML.ReadTag;
  var
    S: TString;
begin
  Result := False;
  XMLSetEmpty(Item, xitTag, '');
  SkipSpaces(Data);
  Item.Text := '';
  while not (TChar(Data.GetPtr^) in [#0..#32, '>', '/']) do
    Item.Text := Item.Text + TChar(Data.ReadByte);
  if not ReadParams(Data, Item, ['/', '>']) then
    Exit;
  case Data.ReadByte of
    Ord('/'): 
      if Data.ReadByte <> Ord('>') then
        Exit;
    Ord('>'): 
      begin
        if not ReadContent(Data, Item, S) then
           Exit;
        if S <> Item.Text then
          Exit;
      end;
  else
    Exit;
  end;
  Result := True;
end;

function TXML.ReadContent(Data: TData; Item: PXMLItem; var EndTag: TString): Boolean;
  var
    Text: TString;
begin
  Result := False;
  SkipSpaces(Data);
  while True do begin
    Text := '';
    while TChar(Data.GetPtr^) <> '<' do
      Text := Text + TChar(Data.ReadByte);
    if Text <> '' then
      XMLAddContent(Item, xitText, Text);
    Data.ReadByte;
    if TChar(Data.GetPtr^) <> '/' then begin 
      if not ReadTag(Data, XMLAddContent(Item, xitTag, '')) then
        Exit
    end else begin
      Text := '';
      Data.ReadByte;
      SkipSpaces(Data);
      while not (TChar(Data.GetPtr^) in [#0..#32, '>']) do
        Text := Text + TChar(Data.ReadByte);
      EndTag := Text;
      SkipSpaces(Data);
      if Data.ReadByte <> Ord('>') then
        Exit;
      Break;
    end;
    SkipSpaces(Data);
  end;
  Result := True;
end;

function TXML.ReadStr(Data: TData; var S: TString): Boolean;
begin
  S := '';
  while Byte(Data.GetPtr^) > 32 do
    S := S + TChar(Data.ReadByte);
  Result := True;
end;

function TXML.GetHeader: PXMLItem;
begin
  Result := @FHeader;
end;

function TXML.GetRoot: PXMLItem;
begin
  Result := @FRoot;
end;

end.
