unit e_particles;

interface

uses OpenGL, e_types, e_utils, e_models, e_textures, e_font, e_screen, e_global;

const
  particles_smoke           = 1;
  particles_explosion       = 2;
  particles_shoot           = 3;
  particles_bonus           = 4;

type
  TParticle = record
    pos                    : TPoint;
    speed                  : TPoint;
    angle                  : single;
    life                   : single;
    start_life             : single;
    size                   : single;
    color                  : TColor;
  end;

  TParticlesEmitter = class
    objects : array of TParticle;
    pos     : TPoint;
    active  : boolean;
    color   : TColor;
    texture : integer;
    layer   : byte;
    loop    : boolean;
    multiply : boolean;
    parent  : TCreature;
    size    : single;
  public
    constructor Create(p : TPoint; n : integer; c : TColor);   virtual;
    procedure   ResetParticle(i : integer);                    virtual;
    procedure   UpdateParticle(i : integer; count : single);   virtual;
    procedure   Update(count : single);                        virtual;
    procedure   Render;                                        virtual;
  end;

  TParticlesSmoke = class(TParticlesEmitter)
  public
    constructor Create(p : TPoint; n : integer; c : TColor);   override;
    procedure   ResetParticle(i : integer);                    override;
    procedure   UpdateParticle(i : integer; count : single);   override;
  end;

  TParticlesShoot = class(TParticlesEmitter)
  public
    constructor Create(p : TPoint; n : integer; c : TColor);   override;
    procedure   ResetParticle(i : integer);                    override;
    procedure   UpdateParticle(i : integer; count : single);   override;
  end;

  TParticlesBonus = class(TParticlesEmitter)
  public
    constructor Create(p : TPoint; n : integer; c : TColor);   override;
    procedure   ResetParticle(i : integer);                    override;
    procedure   UpdateParticle(i : integer; count : single);   override;
  end;

  TParticlesExplosion = class(TParticlesEmitter)
  public
    constructor Create(p : TPoint; n : integer; c : TColor);   override;
    procedure   ResetParticle(i : integer);                    override;
    procedure   UpdateParticle(i : integer; count : single);   override;
  end;

  TParticlesEngine = class
    objects : array of TParticlesEmitter;
  public
    constructor Create;
    function    AddEmitter(m : word; p : TPoint; n : integer; c : TColor) : TParticlesEmitter;
    procedure   DeleteEmitter(i : integer);
    procedure   Update(count : single);
    procedure   Render(layer : byte);
  end;

var
  particles             : TParticlesEngine;
  particles_active      : boolean = true;
  particles_in_frame    : integer = 0;
  particles_updates     : integer = 0;

implementation

uses e_game, e_camera, e_player;

{-- TParticlesEmitter --}

constructor TParticlesEmitter.Create(p : TPoint; n : integer; c : TColor);
var
i : integer;
a : single;
begin
inherited Create;

pos     := p;
color   := c;
active  := true;
size    := 1;
end;

procedure TParticlesEmitter.ResetParticle(i : integer);
begin
end;

procedure TParticlesEmitter.UpdateParticle(i : integer; count : single);
begin
end;

procedure TParticlesEmitter.Update(count : single);
var
i, j : integer;
begin
j := 0;
for i := 0 to high(objects) do
  with objects[i] do
    if life > 0 then
      begin
        life := life - count;
        UpdateParticle(i, count);
        inc(j);
      end
    else
      if loop then
        ResetParticle(i);
        
if (not loop) and (j <= 0) then active := false;
end;

procedure TParticlesEmitter.Render;
var
i : integer;
s : single;
begin
SetTexture(texture);

if multiply then
  glBlendFunc(GL_SRC_ALPHA, GL_ONE)
else
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  
for i := 0 to high(objects) do
  with objects[i] do
    if life > 0 then
      if CircleVisible(pos, size) then
      begin
        inc(particles_in_frame);
        with color do glColor4f(r, g, b, a);
        glPushMatrix;
        with pos do glTranslatef(x, y, 0);
        glScalef(size, size, size);
        glRotatef(angle, 0, 0, 1);
        glCallList(model_quad);
        glPopMatrix;
      end;

if multiply then glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

end;

{-- TParticlesEmitter --}

{-- TParticlesSmoke --}

constructor TParticlesSmoke.Create(p : TPoint; n : integer; c : TColor);
var
i : integer;
a : single;
begin
inherited Create(p, n, c);

texture := texture_smoke;
layer := 2;
loop := true;
multiply := false;
SetLength(objects, n);
for i := 0 to high(objects) do
  begin
    ResetParticle(i);
    //objects[i].start_life := i / high(objects) * 800;
    //objects[i].life := objects[i].start_life;
  end;
end;

procedure TParticlesSmoke.ResetParticle(i : integer);
begin
objects[i].pos := pos;
objects[i].color := color;
  with objects[i] do
    begin
      pos.x := pos.x;
      pos.y := pos.y;
      angle := random * 360;
      size  := 0.2;

      start_life  := random * 10 * (high(objects) + 1);
      if start_life > 1000 then
        start_life := 1000;
        
      life        := start_life;
      speed       := GetPoint(GetRandom * 0.1, GetRandom * 0.1);
    end;
end;

procedure TParticlesSmoke.UpdateParticle(i : integer; count : single);
var
s : single;
p : TPoint;
begin
if parent <> nil then
  pos := parent.pos;
p := pos;
s := 1.5;
if multiply then
  s := s * 0.75;

with objects[i] do
  begin
    pos.x := pos.x + speed.x * count * 0.001;
    pos.y := pos.y + speed.y * count * 0.001;

    Aspire(size, parent.size * life / start_life, count * 0.005);
    if size < 0 then
      size := 0;
    if size > 1 then
      size := 1;
    
    color.a := life / start_life;
  end;
end;

{-- TParticlesFire --}

{-- TParticlesShoot --}

constructor TParticlesShoot.Create(p : TPoint; n : integer; c : TColor);
var
i : integer;
a : single;
begin
inherited Create(p, n, c);

texture := texture_smoke;
layer := 4;
loop := false;
multiply := false;
SetLength(objects, n);
for i := 0 to high(objects) do
  ResetParticle(i);
end;

procedure TParticlesShoot.ResetParticle(i : integer);
begin
objects[i].pos := pos;
objects[i].color := color;
  with objects[i] do
    begin
      pos.x := pos.x;
      pos.y := pos.y;
      angle := random * 360;
      size  := 0.5 * (1 + random);

      start_life  := 400 + random * 400;
      life        := start_life;
      speed       := GetPoint(cos(angle / 180 * pi) * size, sin(angle / 180 * pi) * size);
      pos.x := pos.x + speed.x * 0.1;
      pos.y := pos.y + speed.y * 0.1;
    end;
end;

procedure TParticlesShoot.UpdateParticle(i : integer; count : single);
var
s : single;
p : TPoint;
begin
with objects[i] do
  begin
    AspirePoint(speed, GetPoint(0, 0), count * 0.0001);
    pos.x := pos.x + speed.x * count * 0.001;
    pos.y := pos.y + speed.y * count * 0.001;

    color.a := 0.7 * life / start_life;
  end;
end;

{-- TParticlesShoot --}

{-- TParticlesBonus --}

constructor TParticlesBonus.Create(p : TPoint; n : integer; c : TColor);
var
i : integer;
a : single;
begin
inherited Create(p, n, c);

texture := texture_bonus;
layer := 4;
loop := false;
multiply := true;
SetLength(objects, n);
for i := 0 to high(objects) do
  ResetParticle(i);
end;

procedure TParticlesBonus.ResetParticle(i : integer);
begin
objects[i].pos := pos;
objects[i].color := color;
  with objects[i] do
    begin
      pos.x := pos.x;
      pos.y := pos.y;
      angle := i / (high(objects) + 1) * 360;
      size  := 0.3;

      start_life  := 400 + random * 400;
      life        := start_life;
      speed       := GetPoint(cos(angle / 180 * pi) * size * 4, sin(angle / 180 * pi) * size * 4);
      pos.x := pos.x + speed.x * 0.1;
      pos.y := pos.y + speed.y * 0.1;
    end;
end;

procedure TParticlesBonus.UpdateParticle(i : integer; count : single);
var
s : single;
p : TPoint;
begin
with objects[i] do
  begin
    AspirePoint(speed, GetPoint(0, 0), count * 0.0005);
    pos.x := pos.x + speed.x * count * 0.001;
    pos.y := pos.y + speed.y * count * 0.001;

    color.a := life / start_life;
  end;
end;

{-- TParticlesBonus --}

{-- TParticlesExplosion --}

constructor TParticlesExplosion.Create(p : TPoint; n : integer; c : TColor);
var
i : integer;
a : single;
begin
inherited Create(p, n, c);

texture := texture_explosion;
layer := 4;
loop := false;
multiply := true;
SetLength(objects, n);
for i := 0 to high(objects) do
  ResetParticle(i);
end;

procedure TParticlesExplosion.ResetParticle(i : integer);
begin
objects[i].pos := pos;
objects[i].color := color;
  with objects[i] do
    begin
      pos.x := pos.x;
      pos.y := pos.y;
      angle := random * 360;
      size  := (1 + random);

      start_life  := 200 + random * 200;
      life        := start_life;
      speed       := GetPoint(cos(angle / 180 * pi) * size * 2, sin(angle / 180 * pi) * size * 2);
    end;
end;

procedure TParticlesExplosion.UpdateParticle(i : integer; count : single);
var
s : single;
p : TPoint;
begin
with objects[i] do
  begin
    AspirePoint(speed, GetPoint(0, 0), count * 0.005);
    pos.x := pos.x + speed.x * count * 0.001;
    pos.y := pos.y + speed.y * count * 0.001;

    color.a := life / start_life;
  end;
end;

{-- TParticlesExplosion --}

{-- TParticlesEngine --}

constructor TParticlesEngine.Create;
var
i : integer;
x, z : single;
begin
SetLength(objects, 0);
end;

function TParticlesEngine.AddEmitter(m : word; p : TPoint; n : integer; c : TColor) : TParticlesEmitter;
var
i, max : integer;
begin
result := nil;

i   := 0;
max := high(objects);
while (i <= max) and (objects[i] <> nil) do inc(i);
if i > max then SetLength(objects, max + 4);

case m of
  particles_smoke            : objects[i] := TParticlesSmoke.Create(p, n, c);
  particles_explosion        : objects[i] := TParticlesExplosion.Create(p, n, c);
  particles_shoot            : objects[i] := TParticlesShoot.Create(p, n, c);
  particles_bonus            : objects[i] := TParticlesBonus.Create(p, n, c);
  else                         objects[i] := TParticlesExplosion.Create(p, n, c);
end;

result := objects[i];
end;

procedure TParticlesEngine.DeleteEmitter(i : integer);
begin
objects[i].Free;
objects[i] := nil;
end;

procedure TParticlesEngine.Update(count : single);
var
i : integer;
begin
if not particles_active then exit;

particles_updates := 0;
for i := 0 to high(objects) do
  if objects[i] <> nil then
    if objects[i].active then
      begin
        objects[i].Update(count);
        inc(particles_updates);
      end
    else
      DeleteEmitter(i);
end;

procedure TParticlesEngine.Render(layer : byte);
var
i : integer;
begin
if not particles_active then exit;

if layer = 1 then
  particles_in_frame := 0;

for i := 0 to high(objects) do
  if objects[i] <> nil then
    if objects[i].active and (objects[i].layer = layer) then
      objects[i].Render;
end;

{-- TParticlesEngine --}

end.
